# -*- coding: utf-8 -*-

"""
Show images, such as thumbnails or regions extracted from whole-slide images
(WSI; also known as virtual slides).

"""

import numpy as np
from matplotlib import pyplot as plt
import matplotlib.patches as mpatch


class Figure(object):
    """An open *annotated* WSI.

    Parameters
    ----------
    image: PIL Image
        Image to display; a thumbnail or region read from WSI.
    annotation: Numpy 2D array
        Annotation labels for each pixel in the input ``image``.
    color_map: dict {label int: RGB color list}
        Mapping of RGB color values to class labels. Colors are triplet lists
        of the R, G, and B values.

    Attributes
    ----------
    image: PIL Image
        The input ``image``.
    annotation: Numpy 2D array
        The input ``annotation``.
    color_map: dict {label: color}
        Mapping between each label and the color for visualization.

    Examples
    --------
    >>> from wsipre import show
    >>> colors = {0: (0, 0, 0), 2: (1, 0, 0)}
    >>> fig = show.Figure(image=image, annotation=mask, color_map=colors)
    >>> fig.color_map
    {0: (0, 0, 0), 2: (1, 0, 0)}

    >>> fig.show_label_colors(height=5)

    >>> fig.show_image_with_annotation()

    """

    def __init__(self, image, annotation, color_map={0: (0., 0., 0.),
                                                     1: (0., .8, 1.),
                                                     2: (1., .8, 0.),
                                                     3: (1., .4, 0.)}):
        self.image = image
        self.annotation = annotation.astype(np.float32)  # In case it's boolean
        self.color_map = color_map

        # RGBA values should be within 0-1 range
        if all([0 <= channel <= 255 for color in self.color_map.values()
                for channel in color]):
            if any([channel > 1 for color in self.color_map.values()
                    for channel in color]):
                self.color_map = {key: [channel / 255.0 for channel in color]
                                  for (key, color) in self.color_map.items()}
        else:
            raise ValueError(
                'Please make sure color triplets values in "color_map" are ' +
                'either between 0-1 or 0-255 range.')

        # Are there enough colors?
        labels = np.unique(self.annotation)
        labels_not_in_map = [label not in color_map.keys() for label in labels]
        if any(labels_not_in_map):
            raise ValueError(
                'No color provided in "color_map" for label(s) ' +
                f' {list(labels[labels_not_in_map])}.')

    def show_label_colors(self, width=20, height=6, font_color='w',
                          font_size=14):
        """Display a color bar with the colors and overlayed labels.

        Parameters
        ----------
        width: int
            Width of each class rectangle in plot.
        height: int
            Height of each class rectangle in plot.
        font_size: int
            Size of the font used to display class labels.

        Returns
        -------
        Matplotlib image illustrating mapping between class label and color.

        """
        rectangles = [mpatch.Rectangle((0 + width * i, 0), width, height,
                                       color=list(self.color_map.values())[i])
                      for i in range(len(self.color_map.keys()))]
        rectangles = {list(self.color_map.keys())[i]: rect
                      for i, rect in enumerate(rectangles)}

        fig, ax = plt.subplots()

        for r in rectangles:
            ax.add_artist(rectangles[r])
            rx, ry = rectangles[r].get_xy()
            cx = rx + rectangles[r].get_width() / 2.0
            cy = ry + rectangles[r].get_height() / 2.0

            ax.annotate(r, (cx, cy), color=font_color, weight='bold',
                        fontsize=font_size, ha='center', va='center')

        ax.set_xlim((0, width * len(rectangles)))
        ax.set_ylim((0, height))
        ax.set_aspect('equal')
        ax.axis('off')
        plt.show()

    def show_image(self):
        """Show loaded image.

        Display the loaded WSI thumbnail or region.

        Returns
        -------
        Matplotlib image.

        """
        plt.imshow(self.image)
        plt.show()

    def _paint_annotation_mask(self):
        """Convert 2D mask to 3D format (RGB image)."""
        red = self.annotation.copy()
        green = self.annotation.copy()
        blue = self.annotation.copy()

        for label, color in self.color_map.items():
            idx = self.annotation == label
            red[idx], green[idx], blue[idx] = color

        rgb = np.stack([red, green, blue], axis=2)

        return rgb

    def show_annotation(self):
        """Show loaded image annotation.

        Display the loaded WSI thumbnail or region annotation, corresponding to
        a mask of pixel-wise labels.

        Returns
        -------
        Matplotlib image.

        """
        rgb_mask = self._paint_annotation_mask()
        plt.imshow(rgb_mask)
        plt.show()

    def show_image_with_annotation(self, split=True):
        """Show loaded image and annotation.

        Parameters
        ----------
        split: bool
            Whether to display the split image and annotation side-by-side or a
            combined visualization of the annotation on the image.

        Returns
        -------
        Matplotlib image.

        """
        if split:
            rgb_mask = self._paint_annotation_mask()

            f, (ax1, ax2) = plt.subplots(1, 2, sharey=True)
            ax1.imshow(np.array(self.image))
            ax2.imshow(rgb_mask)
            plt.show()
        else:
            image = np.array(self.image)[:, :, :3]
            if np.max(image) > 1:
                image = image / 255.

            for label, color in self.color_map.items():
                if label == 0:
                    continue
                idx = self.annotation == label
                image[idx] = color

            plt.imshow(image)
            plt.show()



# -*- coding: utf-8 -*-

"""
Handle annotated whole-slide images (WSI; also known as virtual slides).

"""

import warnings
from distutils.version import StrictVersion
import random

import numpy as np
import openslide
import cv2

import annotation as reader
import tissue


class _AnnotatedOpenSlide(openslide.OpenSlide):
    """An open *annotated* WSI.

    Wrapper of `openslide.OpenSlide` to handle annotation polygons together
    with their corresponding OpenSlide Python WSI instances.

    Parameters
    ----------
    filename: str
        Path to slide file.
    annotation_filename: str
        Path to XML annotation file.
    xml_style: str
        The annotation XML style. Typically associated with a computational
        histology challenge releasing the dataset.
        * 'asap': CAMELYON grand challenges in pathology.
        * 'bach': BACH Grand Challenge on Breast Cancer Histology Images.

    Attributes
    ----------
    filename: str
        The WSI file name to read.
    annotation_filename: str
        Corresponding XML annotation file.
    polygons: list of lists of 2-tuples
        Polygon region annotations.
    labels: list of str
        Polygon region annotation labels.
    label_map: dict {str: int}
        Correspondence between polygon region annotation labels and integer
        values used in numpy masks.

    """

    def __init__(self, filename, annotation_filename, xml_style):
        openslide.OpenSlide.__init__(self, filename)
        self.filename = filename  # Useful to name predicted annotations
        self.annotation_filename = annotation_filename
        self.xml_style = xml_style
        self.polygons = None
        self.label_map = None

        if self.annotation_filename is not None:
            if self.xml_style == 'asap':
                self.polygons, self.labels = reader.asap_annotations(
                    self.annotation_filename)
                # CAMELYON17 data
                if 'metastases' in self.labels:
                    # Avoid value 0 (used by default for unlabeled regions)
                    self.label_map = {'metastases': 2, 'normal': 1}
                # CAMELYON16 data
                elif '_0' in self.labels or '_1' in self.labels:
                    self.label_map = {'_0': 2, '_1': 2, '_2': 1}
                else:  # Predicted annotations
                    self.label_map = {'predicted_tumor': 1}
            elif self.xml_style == 'bach':
                # Value 1 is reserved for 'normal' tissue annotations
                self.label_map = {'Benign': 2, 'Carcinoma in situ': 3,
                                  'Invasive carcinoma': 4}
                self.polygons, self.labels = reader.bach_annotations(
                    self.annotation_filename)
            else:
                raise ValueError(
                    '"xml_style" value must be either "asap" or "bach".')
        else:
            if self.xml_style is not None:
                warnings.warn('"xml_style" is only used if an ' +
                              '"annotation_filename" is provided.')


class Slide(_AnnotatedOpenSlide):
    """An open WSI, with or without annotations.

    Added functionality to OpenSlide WSI instances, optionally with
    annotations.

    Parameters
    ----------
    filename: str
        Path to slide file. Namesake argument to `openslide.OpenSlide` and
        `AnnotatedOpenSlide`.
    annotation_filename: str
        Path to XML annotation file. Namesake argument to `AnnotatedOpenSlide`.
    xml_style: str
        The computational histology challenge releasing the dataset. Namesake
        argument to `AnnotatedOpenSlide`.

    Attributes
    ----------
    tissue_mask: Numpy 2D array
        Tissue RoI ``annotation``.
    tissue_label_map: dict {str: int}
        Correspondence between polygon region annotation labels and integer
        values used in numpy masks.
    downsampling_factor: float
        The scaling factor relative to level 0 dimensions.
    downsampled_slide: PIL image
        The scaled down slide generated when running the tissue detector.

    Examples
    --------
    >>> from wsipre import slide
    >>> wsi = slide.Slide('tumor_001.tif', 'tumor_001.xml', 'asap')
    >>> wsi.label_map
    {'_0': 2, '_1': 2, '_2': 1}
    >>> wsi.level_count
    10
    >>> len(wsi.polygons)
    2

    >>> thumbnail, mask, dwnspl_factor = wsi.get_thumbnail_with_annotation(
    ...     size=(3000, 3000), polygon_type='line', line_thickness=5)

    >>> slide_region = wsi.read_region_with_annotation(
    ...     location=(65000, 110000), level=4, size=(1000, 1000))

    """

    def __init__(self, filename, annotation_filename=None, xml_style=None):
        self.tissue_mask = None
        self.tissue_label_map = None
        self.downsampling_factor = None
        self.downsampled_slide = None

        # OpenSlide version < 3.4.1 lacks support for some recent file formats
        version = openslide.__library_version__
        if StrictVersion(version) < StrictVersion('3.4.1'):
            warnings.warn(
                f'OpenSlide version {version} lacks support for some ' +
                'whole-slide image file formats. It is highly recommended ' +
                'to update to a more recent version.')

        if annotation_filename is None:
            if xml_style is not None:
                warnings.warn(
                    '"xml_style" is not used (no annotation was provided).')
            openslide.OpenSlide.__init__(self, filename)
        else:
            _AnnotatedOpenSlide.__init__(
                self, filename, annotation_filename, xml_style)

    def _draw_polygons(self, mask, polygons, polygon_type,
                       line_thickness=None):
        """Convert polygon vertex coordinates to polygon drawing."""
        for poly, label in zip(polygons, self.labels):
            if polygon_type == 'line':
                mask = cv2.polylines(mask, [poly], True,
                                     int(self.label_map[label]),
                                     line_thickness)
            elif polygon_type == 'area':
                if line_thickness is not None:
                    warnings.warn('"line_thickness" is only used if ' +
                                  '"polygon_type" is "line".')

                mask = cv2.fillPoly(mask, [poly], int(self.label_map[label]))
            else:
                raise ValueError(
                    'Accepted "polygon_type" values are "line" or "area".')

        return mask

    def _draw_tissue_polygons(self, mask, polygons, polygon_type,
                              line_thickness=None):
        """Convert tissue polygon vertex coordinates to polygon drawing."""
        tissue_label = 1

        for poly in polygons:
            if polygon_type == 'line':
                mask = cv2.polylines(
                    mask, [poly], True, tissue_label, line_thickness)
            elif polygon_type == 'area':
                if line_thickness is not None:
                    warnings.warn('"line_thickness" is only used if ' +
                                  '"polygon_type" is "line".')

                mask = cv2.fillPoly(mask, [poly], tissue_label)
            else:
                raise ValueError(
                    'Accepted "polygon_type" values are "line" or "area".')

        return mask

    def get_thumbnail_with_annotation(self, size, polygon_type='area',
                                      line_thickness=None):
        """Convert *annotated* WSI to a scaled-down thumbnail.

        Parameters
        ----------
        size: 2-tuple
            Maximum size of compressed image thumbnail as (width, height).
        polygon_type: str
            Type of polygon drawing:
            * 'line'
            * 'area'
        line_thickness: int
            Polygon edge line thickness. Required if ``polygon_type`` is
            ``line``.

        Returns
        -------
        Scaled down slide (as PIL RGBA image) and annotation mask (as numpy 2D
        array), plus the scaling factor.

        """
        if self.polygons is None:
                raise ValueError(
                    'No annotation is available. Please load an annotation ' +
                    'or consider using "get_thumbnail" instead.')

        thumb = self.get_thumbnail(size)

        width_factor = self.dimensions[0] / size[0]
        height_factor = self.dimensions[1] / size[1]
        downsampling_factor = max(width_factor, height_factor)

        polygons = np.array([np.array(poly) for poly in self.polygons])
        polygons = polygons / downsampling_factor
        polygons = [poly.astype('int32') for poly in polygons]

        mask = np.zeros(thumb.size[::-1])  # Reverse width and height
        mask = self._draw_polygons(
            mask, polygons, polygon_type, line_thickness)

        return thumb, mask, downsampling_factor

    def read_region_with_annotation(self, location, level, size,
                                    polygon_type='area', line_thickness=None):
        """Crop a smaller region from an *annotated* WSI.

        Get a defined region from a WSI, together with its annotation mask.

        Parameters
        ----------
        location: 2-tuple
            X and y coordinates of the top left pixel. Namesake argument to
            OpenSlide's "read_region" function.
        level: int
            Slide level. Namesake argument to OpenSlide's "read_region"
            function.
        size: 2-tuple
            Crop size (width, height). Namesake argument to OpenSlide's
            "read_region" function.
        polygon_type: str
            Type of polygon drawing:
            * 'line'
            * 'area'
        line_thickness: int
            Polygon edge line thickness. Required if 'polygon_type' is 'line'.

        Returns
        -------
        Image region (as PIL RGBA image) and the annotation mask (as numpy 2d
        array).

        """
        if self.polygons is None:
                raise ValueError(
                    'No annotation is available. Please load an annotation ' +
                    'or consider using "read_region" instead.')
        downsampling_factor = self.level_downsamples[level]
        slide_region = self.read_region(location, level, size)

        polygons = np.array([np.array(poly) for poly in self.polygons])
        polygons = polygons / downsampling_factor
        polygons = [poly.astype('int32') for poly in polygons]

        location = tuple(int(coord / downsampling_factor)
                         for coord in location)

        # Convert polygon coordinates to patch coordinates
        polygons = [poly - np.array(location) for poly in polygons]

        mask_region = np.zeros(size[::-1])
        mask_region = self._draw_polygons(
           mask_region, polygons, polygon_type, line_thickness)

        return slide_region, mask_region

    def get_tissue_mask(self, downsampling_factor=64, polygon_type='area',
                        line_thickness=None):
        """Generate tissue region annotation.

        Make a binary mask annotating tissue regions of interest (RoI) on the
        WSI using an automatic threshold-based segmentation method inspired by
        the one used by Wang *et al.* [1]_. Briefly, the method consists on the
        following steps, starting from a WSI:
            * Select downsampling level (typically a factor of 64)
            * Transfer from the RGB to the HSV color space
            * Determine optimal threshold value in the saturation channel using
              the Otsu algorithm [2]_
            * Threshold image to generate a binary mask
            * Fill in small holes and remove small objects

        .. [1] Dayong Wang, Aditya Khosla, Rishab Gargeya, Humayun Irshad,
           Andrew H. Beck, "Deep Learning for Identifying Metastatic Breast
           Cancer", arXiv:1606.05718.

        .. [2] Nobuyuki Otsu, "A Threshold Selection Method from Gray-level
           Histograms", IEEE Trans Syst Man Cybern., 9(1):62–66, 1979.

        Parameters
        ----------
        downsampling_factor: int
            The desired factor to downsample the image by, since full WSIs will
            not fit in memory. The image's closest level downsample is found
            and used.
        polygon_type: str
            Type of polygon drawing:
            * 'line'
            * 'area'
        line_thickness: int
            Polygon edge line thickness. Required if ``polygon_type`` is
            ``line``.

        Returns
        -------
        Binary mask as numpy 2D array, RGB slide image (in the used
        downsampling level, to allow visualization) and downsampling factor.

        """
        (mask_contours,
         self.downsampled_slide,
         self.downsampling_factor) = tissue.detect_tissue(
             self, downsampling_factor)

        mask = np.zeros(self.downsampled_slide.shape[:2])
        self.tissue_mask = self._draw_tissue_polygons(
           mask, mask_contours, polygon_type, line_thickness)
        self.tissue_label_map = {'background': 0, 'tissue': 1}

        return self

    def _get_random_coordinates(self, pixels_in_roi, level,
                                mask_downsampling_factor, size):
        """Get region (patch) coordinates at random."""
        coordinates = random.choice(pixels_in_roi)

        # Scale coordinates up to level-0 dimensions
        coordinates = [int(x * mask_downsampling_factor) for x in coordinates]

        # OpenSlide Python's read_region takes top left corner position, which
        # effectively excludes tissue above and to the left of annotations
        # Fix it by offsetting coordinates: convert to center pixel position
        # (subtract half of patch size from x and y coordinates)
        half_size = [x * self.level_downsamples[level] / 2 for x in size]
        row_location = coordinates[0] - half_size[0]
        col_location = coordinates[1] - half_size[1]

        # Make sure coordinates are still within slide margins
        cols_edge = self.dimensions[0] - size[0]
        rows_edge = self.dimensions[1] - size[1]
        row_location = max(0, min(row_location, rows_edge))
        col_location = max(0, min(col_location, cols_edge))

        return int(col_location), int(row_location)  # OpenSlide: width, height

    def _get_pixels_avoiding_labels(self, avoid_labels):
        # Can only do this if an annotation is available
        if self.polygons is None:
            raise ValueError('No annotation is available.')

        # Generate annotated thumbnail with exact dimensions of tissue mask
        rows, cols = self.tissue_mask.shape
        _, mask, _ = self.get_thumbnail_with_annotation(
            size=(cols, rows), polygon_type='area')

        # Locate pixels in tissue and not in "avoid_labels"
        new_mask = np.where(self.tissue_mask == 1, 1, mask)
        new_mask = np.where(np.isin(mask, avoid_labels), 2, new_mask)

        return tuple(zip(*np.where(new_mask == 1)))

    def _max_repeated_pixel_ratio(self, image):
        """Compute ratio of count of most common pixel value to total count."""
        image = np.array(image)
        _, counts = np.unique(image, return_counts=True)

        return np.max(counts) / image.size

    def _get_area_ratio(self, mask, target_class):
        """Compute ratio of pixels labeled as 'target_class' to all pixels."""
        class_pixels = mask[np.where(mask == target_class)].size

        return class_pixels / mask.size

    def read_random_tissue_patch(self, level, size, avoid_labels=None):
        """Crop random patch from detected tissue on WSI.

        Parameters
        ----------
        level: int
            Slide level. Namesake argument to OpenSlide's "read_region"
            function.
        size: 2-tuple
            Crop size (width, height). Namesake argument to OpenSlide's
            "read_region" function.
        avoid_labels: iterable
            Labels from provided annotation to avoid when targeting tissue.
            An annotation is must be available.

        Returns
        -------
        Image region or patch as PIL RGBA image (plus annotation mask as 2D
        Numpy array if labels to avoid are provided).

        """
        if self.tissue_mask is None:
            self.get_tissue_mask()

        if avoid_labels is None:
            # Select coordinates matching tissue (labeled as 1)
            pixels_in_roi = tuple(zip(*np.where(self.tissue_mask == 1)))
        else:
            # Select coordinates matching tissue (labeled as 1) and avoiding
            # indicated labels
            # Can only do this if an annotation is available
            if self.polygons is None:
                raise ValueError('No annotation is available.')

            # Generate annotated thumbnail with exact dimensions of tissue mask
            rows, cols = self.tissue_mask.shape
            _, mask, _ = self.get_thumbnail_with_annotation(
                size=(cols, rows), polygon_type='area')

            # Locate pixels in tissue and not in "avoid_labels"
            new_mask = np.where(self.tissue_mask == 1, 1, mask)
            new_mask = np.where(np.isin(mask, avoid_labels), 2, new_mask)

            pixels_in_roi = tuple(zip(*np.where(new_mask == 1)))

        coordinates = self._get_random_coordinates(
            pixels_in_roi, level, self.downsampling_factor, size)
        patch = self.read_region(coordinates, level, size)

        # Avoid false positive regions where a large proportion of pixels is
        # exactly the same
        while self._max_repeated_pixel_ratio(patch) > 0.5:
            coordinates = self._get_random_coordinates(
                pixels_in_roi, level, self.downsampling_factor, size)
            patch = self.read_region(coordinates, level, size)

        # Also, make sure the patch is mostly normal tissue
        # (or backgound, not actually excluding it here...)
        if avoid_labels is not None:
            patch, mask = self.read_region_with_annotation(
                coordinates, level, size, polygon_type='area')
            avoid_ratios = [self._get_area_ratio(mask, x)
                            for x in avoid_labels]
            i = 0
            while any(x > 0.1 for x in avoid_ratios):
                i += 1
                coordinates = self._get_random_coordinates(
                    pixels_in_roi, level, self.downsampling_factor, size)
                patch, mask = self.read_region_with_annotation(
                    coordinates, level, size, polygon_type='area')
                avoid_ratios = [self._get_area_ratio(mask, x)
                                for x in avoid_labels]
                if i > 15:
                    raise ValueError(
                        'Cannot seem to find patch matching requirements.')

            return patch, mask

        return patch

    def _pick_random_coordinates(self, level, size, target_class):
        """Pick random patch to crop from WSI."""
        # Get thumbnail and select coordinates at random
        _, mask, downsampling_factor = self.get_thumbnail_with_annotation(
             size=(5000, 5000), polygon_type='area')

        # Select coordinates matching target class
        pixels_in_roi = tuple(zip(*np.where(mask == target_class)))

        coordinates = self._get_random_coordinates(
            pixels_in_roi, level, downsampling_factor, size)

        # Get ratio (polygon type must be 'area')
        _, mask = self.read_region_with_annotation(
            coordinates, level, size, polygon_type='area')
        area_ratio = self._get_area_ratio(mask, target_class)

        return coordinates, area_ratio

    def read_random_patch(self, level, size, target_class,
                          min_class_area_ratio, polygon_type='area',
                          line_thickness=None):
        """Crop random patch from WSI.

        Select random location within target class according to provided
        annotation.

        Parameters
        ----------
        level: int
            Slide level. Namesake argument to OpenSlide's "read_region"
            function.
        size: 2-tuple
            Crop size (width, height). Namesake argument to OpenSlide's
            "read_region" function.
        target_class: int
            The class annotation of the central pixel of the patch.
            function.
        min_class_area_ratio: float (0, 1]
            Minimum ratio of target class pixels to total pixels.
        polygon_type: str
            Type of polygon drawing:
            * 'line'
            * 'area'
        line_thickness: int
            Polygon edge line thickness. Required if ``polygon_type`` is
            ``line``.

        Returns
        -------
        Image region or patch as PIL RGBA image plus annotation mask as 2D
        Numpy array.

        """
        if not 0 < min_class_area_ratio <= 1:
            raise ValueError(
                '"min_class_area_ratio" must be in the interval (0, 1].')

        coordinates, area_ratio = self._pick_random_coordinates(
            level, size, target_class)

        # When the slide level is high, the following while loop may be
        # infinite, since the tumor area will never be high enough.
        # Try to sample a few times and throw error if not successful
        i = 0
        while area_ratio < min_class_area_ratio:
            i += 1
            coordinates, area_ratio = self._pick_random_coordinates(
                level, size, target_class)
            if i > 15:
                raise ValueError(
                    'Cannot seem to find patch with a minimum of ' +
                    f'{min_class_area_ratio * 100}% of class {target_class}.' +
                    ' The chosen slide level may be too high.')

        # Get actual data (chosen polygon type)
        patch, mask = self.read_region_with_annotation(
            coordinates, level, size, polygon_type, line_thickness)

        return patch, mask



# -*- coding: utf-8 -*-

"""
wsipre.annotation_reader
~~~~~~~~~~~~~~~~~~~~~~~~

Functionality to read whole-slide image (WSI) annotations (from XML files).
"""

from xml.dom import minidom


def asap_annotations(xml_file):
    """Get annotations from XML file in the ASAP style.

    Collects region annotations bounded by polygonal contours from an XML
    annotation file in the ASAP style (popularized by the CAMELYON challenge),
    along with their labels.

    Parameters
    ----------
    xml_file: str
        Path to XML file.
    Returns
    -------
    List of regions (lists of 2-tuples of x, y polygon vertex coordinates) and
    list of region labels.
    """
    xml = minidom.parse(xml_file)
    annotations = xml.getElementsByTagName('Annotation')

    polygons, labels = [], []
    for annotation in annotations:
        label = annotation.attributes['PartOfGroup'].value
        if label == 'None':
            continue

        labels.append(label)
        vertices = annotation.getElementsByTagName('Coordinate')

        polygon = []
        for vertex in vertices:
            polygon.append((float(vertex.attributes['X'].value),
                            float(vertex.attributes['Y'].value)))
        polygons.append(polygon)

    return polygons, labels


def bach_annotations(xml_file):
    """Get annotations in XML file from the BACH ICIAR 2018 challenge.

    Collects region annotations bounded by polygonal contours from an XML
    annotation file, along with their labels.

    Parameters
    ----------
    xml_file: str
        Path to XML file.
    Returns
    -------
    List of regions (lists of 2-tuples of x, y polygon vertex coordinates) and
    list of region labels.
    """
    xml = minidom.parse(xml_file)
    annotations = xml.getElementsByTagName('Region')

    polygons, labels = [], []
    for annotation in annotations:
        vertices = annotation.getElementsByTagName('Vertex')
        label = annotation.getElementsByTagName('Attribute')
        if label:
            labels.append(label[0].attributes['Value'].value)
        else:
            labels.append(annotation.getAttribute('Text'))

        polygon = []
        for vertex in vertices:
            polygon.append((float(vertex.attributes['X'].value),
                            float(vertex.attributes['Y'].value)))
        polygons.append(polygon)

    return polygons, labels



# -*- coding: utf-8 -*-

"""
wsipre.tissue
-------------

Functionality to detect tissue RoIs.
"""

import numpy as np
import cv2
from skimage import morphology


def otsu_filter(channel, gaussian_blur=True):
    """Otsu filter."""
    if gaussian_blur:
        channel = cv2.GaussianBlur(channel, (5, 5), 0)
    channel = channel.reshape((channel.shape[0], channel.shape[1]))

    return cv2.threshold(
        channel, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]


def detect_tissue(wsi, downsampling_factor=64):
    """Find RoIs containing tissue in WSI.

    Generate mask locating tissue in an WSI. Inspired by method used by
    Wang et al. [1]_.

    .. [1] Dayong Wang, Aditya Khosla, Rishab Gargeya, Humayun Irshad, Andrew
    H. Beck, "Deep Learning for Identifying Metastatic Breast Cancer",
    arXiv:1606.05718

    Parameters
    ----------
    wsi: OpenSlide/AnnotatedOpenSlide class instance
        The whole-slide image (WSI) to detect tissue in.
    downsampling_factor: int
        The desired factor to downsample the image by, since full WSIs will
        not fit in memory. The image's closest level downsample is found
        and used.

    Returns
    -------
    Binary mask as numpy 2D array, RGB slide image (in the used
    downsampling level, in case the user is visualizing output examples)
    and downsampling factor.

    """
    # Get a downsample of the whole slide image (to fit in memory)
    downsampling_factor = min(
        wsi.level_downsamples, key=lambda x: abs(x - downsampling_factor))
    level = wsi.level_downsamples.index(downsampling_factor)

    slide = wsi.read_region((0, 0), level, wsi.level_dimensions[level])
    slide = np.array(slide)[:, :, :3]

    # Convert from RGB to HSV color space
    slide_hsv = cv2.cvtColor(slide, cv2.COLOR_BGR2HSV)

    # Compute optimal threshold values in each channel using Otsu algorithm
    _, saturation, _ = np.split(slide_hsv, 3, axis=2)

    mask = otsu_filter(saturation, gaussian_blur=True)

    # Make mask boolean
    mask = mask != 0

    mask = morphology.remove_small_holes(mask, area_threshold=5000)
    mask = morphology.remove_small_objects(mask, min_size=5000)

    mask = mask.astype(np.uint8)
    _, mask_contours, _ = cv2.findContours(
        mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

    return mask_contours, slide, downsampling_factor
